#pragma once

#include <stdint.h>
#include "i2c/memory.hpp"
#include "../utils/date_time.hpp"
#include "../utils/bcd.hpp"


namespace Devices
{
    template <class I2cImpl>
    class Mcp7940n
    {
        private:

        public:

            using Base = I2c::Memory<I2cImpl>;

            static uint8_t const ADDRESS = 0xde;

            static uint8_t const REG_ADDR_RTCSEC = 0x00;
            static uint8_t const ST = 7;

            static uint8_t const REG_ADDR_RTCMIN = 0x01;
            static uint8_t const REG_ADDR_RTCHOUR = 0x02;
            static uint8_t const FORMAT_12_24 = 6;

            static uint8_t const REG_ADDR_RTCWKDAY = 0x03;
            static uint8_t const OSCRUN = 5;
            static uint8_t const PWRFAIL = 4;
            static uint8_t const VBATEN = 3;

            static uint8_t const REG_ADDR_RTCDATE = 0x04;
            static uint8_t const REG_ADDR_RTCMTH = 0x05;
            static uint8_t const REG_ADDR_RTCYEAR = 0x06;

            static uint8_t const REG_ADDR_CONTROL = 0x07;
            static uint8_t const OUT = 7;
            static uint8_t const SQWEN = 6;
            static uint8_t const ALM1EN = 5;
            static uint8_t const ALM0EN = 4;
            static uint8_t const EXTOSC = 3;
            static uint8_t const CRSTRIM = 2;
            static uint8_t const SQWFS1 = 1;
            static uint8_t const SQWFS0 = 0;

            static uint8_t const REG_ADDR_OSCTRIM = 0x08;

            static uint8_t const REG_ADDR_ALM0SEC = 0x0a;
            static uint8_t const REG_ADDR_ALM0MIN = 0x0b;
            static uint8_t const REG_ADDR_ALM0HOUR = 0x0c;
            static uint8_t const REG_ADDR_ALM0WKDAY = 0x0d;
            static uint8_t const ALM0POL = 7;
            static uint8_t const ALM0MSK2 = 6;
            static uint8_t const ALM0MSK1 = 5;
            static uint8_t const ALM0MSK0 = 4;
            static uint8_t const ALM0IF = 3;

            static uint8_t const REG_ADDR_ALM0DATE = 0x0e;
            static uint8_t const REG_ADDR_ALM0MTH = 0x0f;

            static uint8_t const REG_ADDR_ALM1SEC = 0x11;
            static uint8_t const REG_ADDR_ALM1MIN = 0x12;
            static uint8_t const REG_ADDR_ALM1HOUR = 0x13;
            static uint8_t const REG_ADDR_ALM1WKDAY = 0x14;            
            static uint8_t const ALM1POL = 7;
            static uint8_t const ALM1MSK2 = 6;
            static uint8_t const ALM1MSK1 = 5;
            static uint8_t const ALM1MSK0 = 4;
            static uint8_t const ALM1IF = 3;

            static uint8_t const REG_ADDR_ALM1DATE = 0x15;
            static uint8_t const REG_ADDR_ALM1MTH = 0x16;

            static uint8_t const REG_ADDR_PWRDNMIN = 0x18;
            static uint8_t const REG_ADDR_PWRDNHOUR = 0x19;
            static uint8_t const REG_ADDR_PWRDNDATE = 0x1a;
            static uint8_t const REG_ADDR_PWRDNMTH = 0x1b;

            static uint8_t const REG_ADDR_PWRUPMIN = 0x1c;
            static uint8_t const REG_ADDR_PWRUPHOUR = 0x1d;
            static uint8_t const REG_ADDR_PWRUPDATE = 0x1e;
            static uint8_t const REG_ADDR_PWRUPMTH = 0x1f;

            enum class ClockSource
            {
                Disabled,
                External,
                Crystal
            };

            enum class SquareWave 
            {
                Wave1Hz = 0,
                Wave4096Hz = _BV(SQWFS0),
                Wave8192Hz = _BV(SQWFS1),
                Wave32768Hz = _BV(SQWFS1) | _BV(SQWFS0),
                WaveDisabled
            };

            enum class AlarmMatch
            {
                Seconds = 0x00,
                Minutes = 0x01,
                Hours = 0x02,
                DayOfWeek = 0x03,
                Date = 0x04,
                Time = 0x07
            };

            static bool Init()
            {
                Base::Init();
                return true;
            }

            static bool ConfigureClock(ClockSource const clockSource, bool backupEnabled)
            {
                if (!Base::SetBit(ADDRESS, REG_ADDR_RTCSEC, ST, clockSource == ClockSource::Crystal))
                {
                    return false;
                }

                if (!Base::SetBit(ADDRESS, REG_ADDR_CONTROL, EXTOSC, clockSource == ClockSource::External))
                {
                    return false;
                }

                if (!Base::SetBit(ADDRESS, REG_ADDR_RTCWKDAY, VBATEN, backupEnabled))
                {
                    return false;
                }

                return true;
            }

            static bool SetSquareWave(SquareWave const squareWave)
            {
                uint8_t control;

                if (!Base::Read(ADDRESS, REG_ADDR_CONTROL, control))
                {
                    return false;
                }
                
                if (squareWave == SquareWave::WaveDisabled)
                {
                    control &= ~_BV(SQWEN);
                }
                else
                {
                    control &= ~(_BV(SQWFS1) | _BV(SQWFS0));
                    control |= _BV(SQWEN) | (uint8_t) squareWave;
                }

                return Base::Write(ADDRESS, REG_ADDR_CONTROL, control);
            };

            static bool SetHourFormat(bool const is12Hour)
            {
                return Base::SetBit(ADDRESS, REG_ADDR_RTCHOUR, FORMAT_12_24, is12Hour);
            }

            static bool SetOscillatorTrimmming(int8_t const ppm)
            {
                uint8_t osctrim;
                
                if (ppm > 0) {
                    osctrim = (uint8_t) 0x80 | ppm;
                }
                else if (ppm == -128)
                {
                    osctrim = 0x7f;
                }
                else
                {
                    osctrim = -ppm;
                }

                return Base::Write(ADDRESS, REG_ADDR_OSCTRIM, osctrim);
            };

            static bool GetOscillatorTrimmming(int8_t & ppm)
            {
                uint8_t osctrim;

                if (!Base::Read(ADDRESS, REG_ADDR_OSCTRIM, osctrim))
                {
                    ppm = 0;
                    return false;
                }

                if ((osctrim & 0x80) != 0)
                {
                    ppm = osctrim & 0x7f;
                }
                else
                {
                    ppm = -osctrim;
                }

                return true;
            };

            static bool SetAlarm0(AlarmMatch const alarmMatch, uint8_t time, bool activeState)
            {
                bool result = true;
                uint8_t reg;

                result = result && Base::SetBit(ADDRESS, REG_ADDR_CONTROL, ALM0EN, false);
                // Clear interrupt flag
                result = result && Base::SetBit(ADDRESS, REG_ADDR_ALM0WKDAY, ALM0IF, true);

                reg = Utils::Bcd::IntToBcd(time);

                switch (alarmMatch)
                {
                    case AlarmMatch::Seconds:
                        result = result && Base::Write(ADDRESS, REG_ADDR_ALM0SEC, reg);
                        break;

                    case AlarmMatch::Minutes:
                        result = result && Base::Write(ADDRESS, REG_ADDR_ALM0MIN, reg);
                        break;

                    case AlarmMatch::Hours:
                        result = result && Base::Write(ADDRESS, REG_ADDR_ALM0HOUR, reg);
                        break;

                    case AlarmMatch::Date:
                        result = result && Base::Write(ADDRESS, REG_ADDR_ALM0DATE, reg);
                        break;

                    case AlarmMatch::DayOfWeek:
                        result = result && Base::Write(ADDRESS, REG_ADDR_ALM0WKDAY, reg);
                        break;

                    case AlarmMatch::Time:
                        // Not supported by this method
                        break;
                }

                result = result && Base::Read(ADDRESS, REG_ADDR_ALM0WKDAY, reg);
                reg &= ~_BV(ALM0POL) & ~_BV(ALM0MSK2) & ~_BV(ALM0MSK1) & ~_BV(ALM0MSK0);

                if (activeState)
                {
                    reg |= _BV(ALM0POL);
                }

                reg |= (uint8_t) alarmMatch << ALM0MSK0;
                result = result && Base::Write(ADDRESS, REG_ADDR_ALM0WKDAY, reg);

                result = result && Base::SetBit(ADDRESS, REG_ADDR_CONTROL, ALM0EN, true);
                return result;
            }


            static bool SetAlarm0Seconds(uint8_t const seconds)
            {
                uint8_t reg = Utils::Bcd::IntToBcd(seconds);
                return Base::Write(ADDRESS, REG_ADDR_ALM0SEC, reg);
            }


            static bool ClearAlarm0()
            {
                return Base::SetBit(ADDRESS, REG_ADDR_ALM0WKDAY, ALM0IF, true);
            }


            static bool DisableAlarm0()
            {
                return Base::SetBit(ADDRESS, REG_ADDR_CONTROL, ALM0EN, false);
            }


            static bool ReadDateTime(Utils::DateTime & dateTime)
            {
                uint8_t reg;

                if (!Base::Read(ADDRESS, REG_ADDR_RTCYEAR, reg))
                {
                    return false;
                }

                dateTime.date.year = Utils::Bcd::BcdToInt(reg) + 2000;

                if (!Base::Read(ADDRESS, REG_ADDR_RTCMTH, reg))
                {
                    return false;
                }

                dateTime.date.month = Utils::Bcd::BcdToInt(reg & 0x1f) + 1;

                if (!Base::Read(ADDRESS, REG_ADDR_RTCDATE, reg))
                {
                    return false;
                }

                dateTime.date.day = Utils::Bcd::BcdToInt(reg);
                return true;
            }

            static bool WriteDateTime(Utils::DateTime & dateTime)
            {
                uint8_t reg = Utils::Bcd::IntToBcd(dateTime.date.year - 2000);

                if (!Base::Write(ADDRESS, REG_ADDR_RTCYEAR, reg))
                {
                    return false;
                }

                reg = Utils::Bcd::IntToBcd(dateTime.date.month - 1);

                if (!Base::Write(ADDRESS, REG_ADDR_RTCMTH, reg))
                {
                    return false;
                }

                reg = Utils::Bcd::IntToBcd(dateTime.date.day);

                if (!Base::Write(ADDRESS, REG_ADDR_RTCDATE, reg))
                {
                    return false;
                }

                return true;
            }
    };
}