#include "measurement.hpp"
#include <math.h>
#include "math_utils.hpp"

using namespace Voltmeter;


static constexpr uint8_t const RangeToDecimalPoint(Range const range)
{
    switch (range)
    {
        case Range::_400V:
            return 2;

        case Range::_40V:
            return 1;

        case Range::_4V:
        default:
            return 0;

        case Range::_400mV:
            return 2;
    }
}


static constexpr Voltage::Unit const RangeToUnit(Range const range)
{
    if (range == Range::_400mV)
    {
        return Voltage::Unit::mV;
    }

    return Voltage::Unit::V;
}


int64_t RoundedDivision(int64_t a, int64_t b) 
{
  int64_t q = a / b;
  int64_t r = a % b;
  return q + r / (b / 2 + b % 2);
}


// This assumes that reference voltage is 2.5V and gain is always 0.25*10^x.
static Voltage ValueToVoltage(Range const range, int32_t const value, uint16_t samplesN, uint8_t const digitsN)
{
    uint16_t denominator = MathUtils::Pow10(digitsN);

    return Voltage(
        RangeToUnit(range),
        RoundedDivision((int64_t) value * denominator, (int64_t) samplesN * 1024),
        denominator,
        RangeToDecimalPoint(range)
    );
}


Voltage Measurement::GetMaxVoltage(uint8_t const digitsN, uint32_t calibration)
{
    int32_t calibratedMax = MathUtils::Abs((int32_t) RoundedDivision(calibration, samplesN) - max);
    int32_t calibratedMin = MathUtils::Abs((int32_t) RoundedDivision(calibration, samplesN) - min);

    return ValueToVoltage(range, MathUtils::Max(calibratedMin, calibratedMax), 1, digitsN);
}


Voltage Measurement::GetDcVoltage(uint8_t const digitsN, uint32_t calibration)
{
    return ValueToVoltage(range, (int32_t) calibration - sum, samplesN, digitsN);
}


Voltage Measurement::GetAcVoltage(uint8_t const digitsN)
{
    int64_t rmsAc2 = (sumOfSquares - (int64_t) sum * sum / samplesN) / samplesN;
    int32_t rmsAc = sqrt(rmsAc2);

    uint16_t denominator = MathUtils::Pow10(digitsN);

    return Voltage(
        RangeToUnit(range),
        RoundedDivision((int64_t) rmsAc * denominator, 1024),
        denominator,
        RangeToDecimalPoint(range)
    );
}


uint32_t Measurement::GetCalibration()
{
    return sum;
}